#include <stdio.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <pwd.h>

#include "common.h"
#include "md5.h"
#include "md5util.h"
#include "log.h"

static int _md5_compare(context_t *c, const char *password, const char *ln) {
    md5_state_t st;
    static md5_byte_t digest[16];
    static char t[33];

    if (strlen(ln) != 64) {
        log(c, LOG_WARNING, "Authentication failure: broken MD5 data");
        return PAM_AUTH_ERR;
    }
    
    md5_init(&st);
    md5_append(&st, ln, 32);
    md5_append(&st, password, strlen(password));
    md5_finish(&st, digest);

    fhex_md5(digest, t);
    t[32] = 0;

    return strcmp(ln+32, t) ? PAM_AUTH_ERR : PAM_SUCCESS;
}

static int _check_parent_dirs(const char *fn) {
    static char p[PATH_MAX];
    static struct stat st;
    snprintf(p, sizeof(p), "%s", fn);

    for (;;) {
        char *slash = strrchr(p, '/');

        if (slash == p || !slash)
            return 0;

        *slash = 0;

        if (lstat(p, &st) < 0)
            return -1;

        if (st.st_mode & 022)
            return -1;
    }
}

int user_authentication(context_t *c, const char *username, const char *password) {
    struct passwd *pw;
    FILE *f;
    static char fn[PATH_MAX];
    static struct stat st;
    int ret;
    
    if (!(pw = getpwnam(username))) {
        log(c, LOG_WARNING, "Authentication failure: user <%s> not found", username);
        return PAM_USER_UNKNOWN;
    }

    if (!c->opt_rootok && pw->pw_uid == 0) {
        log(c, LOG_WARNING, "Authentication failure: access denied for root");
        return PAM_AUTH_ERR;
    }

    log(c, LOG_DEBUG, "Searching file for service %s", c->service);

    snprintf(fn, sizeof(fn), "%s/.pam-%s", pw->pw_dir, c->service);
    if (!(f = fopen(fn, "r")) && errno == ENOENT) {
        snprintf(fn, sizeof(fn), "%s/.pam/%s", pw->pw_dir, c->service);
        if (!(f = fopen(fn, "r")) && errno == ENOENT) {
            snprintf(fn, sizeof(fn), "%s/.pam-other", pw->pw_dir);
            if (!(f = fopen(fn, "r")) && errno == ENOENT) {
                snprintf(fn, sizeof(fn), "%s/.pam/other", pw->pw_dir);
                if (!(f = fopen(fn, "r")) && errno == ENOENT) {
                    log(c, LOG_WARNING, "Authentication failure: no .pam file in home directory of <%s> existent", username);
                    return PAM_AUTHINFO_UNAVAIL;
                }
            }
        }
    }

    if (!f) {
        log(c, LOG_WARNING, "Authentication failure: could not open .pam file in home directory of <%s>", username);
        return PAM_AUTH_ERR;
    }

    if (lstat(fn, &st) < 0) {
        log(c, LOG_ERR, "Could not lstat() file %s: %s", fn, strerror(errno));
        fclose(f);
        return PAM_AUTH_ERR;
    }

    if (!S_ISREG(st.st_mode)) {
        log(c, LOG_ERR, "%s ist not a regular file: %s", fn, strerror(errno));
        fclose(f);
        return PAM_AUTH_ERR;
    }
    
    if (fstat(fileno(f), &st) < 0) {
        log(c, LOG_ERR, "Could not fstat() file %s: %s", fn, strerror(errno));
        fclose(f);
        return PAM_AUTH_ERR;
    }

    if (st.st_mode & 0077) {
        log(c, LOG_WARNING, "Authentication failure: bad access mode of file %s: %04o, correct is 0600\n", fn, st.st_mode & 07777);
        fclose(f);
        return PAM_AUTH_ERR;
    }

    if (st.st_uid != pw->pw_uid) {
        log(c, LOG_WARNING, "Authentication failure: bad owner of file %s: %u, correct is %u\n", fn, st.st_uid, pw->pw_uid);
        fclose(f);
        return PAM_AUTH_ERR;
    }

    if (_check_parent_dirs(fn) < 0) {
        log(c, LOG_ERR, "Parent directories of %s must not be group or world writable", fn);
        fclose(f);
        return PAM_AUTH_ERR;        
    }


    ret = PAM_AUTH_ERR;
    while (!feof(f)) {
        static char ln[100];
        int n;

        if (!fgets(ln, sizeof(ln), f))
            break;

        if (ln[0] == 0 || ln[0] == '\n' || ln[0] == '#')
            continue;
        
        if (ln[(n = strlen(ln))-1] == '\n')
            ln[n-1] = 0;

        if (!_md5_compare(c, password, ln)) {
            ret = PAM_SUCCESS;
            break;
        }
    }
    
    fclose(f);

    if (ret == PAM_SUCCESS)
        log(c, LOG_INFO, "Authentication successful for user <%s>", username);
    else
        log(c, LOG_WARNING, "Authentication failure: bad password for user <%s>", username);
    
    return ret;
}
